import sqlalchemy as sa
import typing as t
from sqlalchemy import Boolean, inspect
from sqlalchemy.sql import ColumnElement

from aerie.base import Base
from aerie.collections import Collection
from aerie.queries import SelectQuery
from aerie.session import get_current_session

C = t.TypeVar('C', bound='BaseModel')


class AutoIntegerId:
    __abstract__ = True
    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)


class AutoBigIntegerId:
    __abstract__ = True
    id = sa.Column(sa.BigInteger, primary_key=True, autoincrement=True)


class _QueryProperty(t.Generic[C]):
    def __get__(self, obj: t.Optional[C], type: t.Type[C]) -> SelectQuery[C]:
        return get_current_session().query(type)


class BaseModel(Base):
    __abstract__ = True

    @classmethod
    def query(cls: t.Type[C]) -> SelectQuery[C]:
        return get_current_session().query(cls)

    @classmethod
    async def first(cls: t.Type[C]) -> t.Optional[C]:
        return await cls.query().first()

    @classmethod
    async def all(cls: t.Type[C]) -> Collection[C]:
        return await cls.query().all()

    @classmethod
    async def get(cls: t.Type[C], pk: t.Any, pk_column: str = 'id') -> C:
        column = getattr(cls, pk_column)
        return await cls.query().where(column == pk).one()

    @classmethod
    async def get_or_none(cls: t.Type[C], pk: t.Any, pk_column: str = 'id') -> t.Optional[C]:
        column = getattr(cls, pk_column)
        return await cls.query().where(column == pk).one_or_none()

    @classmethod
    async def get_or_create(
        cls: t.Type[C],
        where: ColumnElement[Boolean],
        values: t.Mapping[str, t.Any],
        autoflush: bool = True,
        autocommit: bool = False,
    ) -> C:
        instance = await cls.query().where(where).one_or_none()
        if not instance:
            instance = await cls.create(autoflush=autoflush, autocommit=autocommit, **values)
        return instance

    @classmethod
    async def create(cls: t.Type[C], autoflush: bool = True, autocommit: bool = False, **values: t.Any) -> C:
        instance = cls(**values)  # type: ignore
        get_current_session().add(instance)
        if autoflush:
            await get_current_session().flush([instance])
        await instance.save(commit=autocommit)  # type: ignore
        return instance

    @classmethod
    async def destroy(cls, *pk: t.Any, pk_column: str = 'id') -> None:
        column = getattr(cls, pk_column)
        for instance in await cls.query().where(column.in_(pk)).all():
            await instance.delete()

    async def save(self, commit: bool = True) -> None:
        session = get_current_session()
        session.add(self)  # type: ignore
        if commit:
            await session.commit()

    async def delete(self, commit: bool = True) -> None:
        session = get_current_session()
        await session.delete(self)
        if commit:
            await session.commit()

    async def refresh(self) -> None:
        session = get_current_session()
        await session.refresh(self)

    def __repr__(self) -> str:
        identity = inspect(self).identity
        if identity is None:
            pk = f"transient {id(self)}"
        elif len(identity) > 1:
            pk = f'pk={identity}'
        else:
            pk = ", ".join(str(value) for value in identity)
            pk = 'pk=' + pk

        return f"<{type(self).__name__}: {pk}>"



from __future__ import annotations

import math
import typing as t

M = t.TypeVar('M')


class Page(t.Generic[M]):
    def __init__(self, rows: t.Sequence[M], total_rows: int, page: int, page_size: int) -> None:
        self.rows = rows
        self.total_rows = total_rows
        self.page = page
        self.page_size = page_size
        self._pointer = 0

    @property
    def total_pages(self) -> int:
        """Total pages in the row set."""
        return math.ceil(self.total_rows / self.page_size)

    @property
    def has_next(self) -> bool:
        """Test if the next page is available."""
        return self.page < self.total_pages

    @property
    def has_previous(self) -> bool:
        """Test if the previous page is available."""
        return self.page > 1

    @property
    def has_other(self) -> bool:
        """Test if page has next or previous pages."""
        return self.has_next or self.has_previous

    @property
    def next_page(self) -> int:
        """Next page number. Always returns an integer.
        If there is no more pages the current page number returned."""
        return min(self.total_pages, self.page + 1)

    @property
    def previous_page(self) -> int:
        """Previous page number. Always returns an integer.
        If there is no previous page, the number 1 returned."""
        return max(1, self.page - 1)

    @property
    def start_index(self) -> int:
        """The 1-based index of the first item on this page."""
        if self.page == 1:
            return 1
        return (self.page - 1) * self.page_size + 1

    @property
    def end_index(self) -> int:
        """The 1-based index of the last item on this page."""
        return min(self.start_index + self.page_size - 1, self.total_rows)

    def iter_pages(
        self, left_edge: int = 3, left_current: int = 3, right_current: int = 3, right_edge: int = 3
    ) -> t.Generator[t.Optional[int], None, None]:
        last = 0
        for number in range(1, self.total_pages + 1):
            if (
                number <= left_edge
                or (number >= self.page - left_current - 1 and number < self.page + right_current)
                or number > self.total_pages - right_edge
            ):
                if last + 1 != number:
                    yield None
                yield number
                last = number

    def __iter__(self) -> t.Iterator[M]:
        return iter(self.rows)

    def __next__(self) -> M:
        if self._pointer == len(self.rows):
            raise StopIteration
        self._pointer += 1
        return self.rows[self._pointer - 1]

    def __getitem__(self, item: int) -> M:
        return self.rows[item]

    def __len__(self) -> int:
        return len(self.rows)

    def __bool__(self) -> bool:
        """A shortcut to check if page has more than one pages.
        Useful in templates to check if the pagination should be rendered or not."""
        return self.total_pages > 1

    def __str__(self) -> str:
        return (
            f'Page {self.page} of {self.total_pages}, rows {self.start_index} - {self.end_index} of {self.total_rows}.'
        )

    def __repr__(self) -> str:
        return f'<Page: page={self.page}, total_pages={self.total_pages}>'



from __future__ import annotations

import typing as t
from sqlalchemy import MetaData
from sqlalchemy.engine import Row
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import Executable

from aerie.base import metadata as shared_metadata
from aerie.results import ResultProxy
from aerie.schema import Schema
from aerie.session import DbSession

_IsolationLevel = t.Literal['SERIALIZABLE', 'REPEATABLE READ', 'READ COMMITTED', 'READ UNCOMMITTED', 'AUTOCOMMIT']

E = t.TypeVar('E', bound=Row)


class Aerie:
    instances: t.Dict[str, Aerie] = {}

    def __init__(
        self,
        url: str,
        echo: bool = False,
        isolation_level: _IsolationLevel = None,
        json_serializer: t.Callable = None,
        json_deserializer: t.Callable = None,
        metadata: MetaData = None,
        name: str = None,
        session_class: t.Type[DbSession] = DbSession,
        session_kwargs: t.Dict[str, t.Any] = None,
        **engine_kwargs: t.Any,
    ) -> None:
        if name is not None:
            if name in Aerie.instances:
                raise KeyError(f'Aerie instance with name "{name}" already exists. Use another name for this instance.')
            Aerie.instances[name] = self

        self.url = url
        self.metadata: MetaData = metadata or shared_metadata
        self.engine: AsyncEngine = create_async_engine(
            url,
            echo=echo,
            isolation_level=isolation_level,
            json_serializer=json_serializer,
            json_deserializer=json_deserializer,
            **engine_kwargs,
        )
        self.schema = Schema(self.engine, self.metadata)

        session_kwargs = session_kwargs or {}
        self._session_maker: sessionmaker = sessionmaker(
            bind=self.engine,
            class_=session_class,
            expire_on_commit=False,
            **session_kwargs,
        )

    def session(self, **options: t.Any) -> t.AsyncContextManager[DbSession]:
        return self._session_maker(**options)

    def transaction(self) -> AsyncEngine._trans_ctx:
        """Establish a new transaction."""
        return self.engine.begin()

    def execute(self, stmt: t.Union[str, Executable], params: t.Mapping = None) -> ResultProxy:
        """Execute an executable or raw SQL query."""
        return ResultProxy(self.engine, stmt, params)

    @classmethod
    def get_instance(cls, name: str = 'default') -> Aerie:
        if name not in Aerie.instances:
            raise KeyError(f'Aerie instance "{name}" does not exists.')
        return Aerie.instances[name]
